# -*- python -*-
# $RCSfile: scriptloader.py,v $
# $Revision: 1.16.8.1 $
# $Author: langer $
# $Date: 2008-08-27 18:57:01 $


# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# ScriptLoader is a derived class of InteractiveInterpreter that
# provides a correct traceback when exceptions occur while processing
# a Python file.  In addition, it provides a hook that can be
# redefined in a derived class to update a progress bar.

# The problem with InteractiveInterpreter is that when an exception is
# raised, the line numbers in the traceback are offsets into the
# current code block, and the current code block is not the file being
# processed.  It is instead just the set of lines passed to runsource
# since the last successful execution.  The traceback uses these
# erroneous line numbers to find the source code for the offending
# instructions.  The filename passed to runsource is only used in
# error messages and to find the source code.  The solution presented
# here is to pass a fake filename to runsource.  The fake name encodes
# the line number of the beginning of the code block.  Then when an
# exception occurs, the actual line numbers can be computed, and the
# traceback can be corrected.

# Because it's difficult to actually create and modify a traceback
# object, this code only creates a modified traceback list, as
# returned by traceback.extract_tb.  That means that the modified
# traceback can't be passed to sys.excepthook.  Instead we pass the
# traceback list to excepthook.printTraceBack().

# Another problem with InteractiveInterpreter is that it doesn't work
# the way execfile() does -- it requires a blank line at the end of an
# indented block, unless the next line is an "else", "elif", or
# "except".  So here we use InteractiveInterpreter to process blocks
# of code, but use the compiler module to find those blocks.  That way
# we're guaranteed that we won't separate an "if" from its "else".
# This is somewhat of a hack, and it fails in some perverse cases.  In
# particular, it won't work if multiline statements are concatentated
# by a semi-colon, like this:
#   f(
#   ); g(
#   )
# This case is sufficiently perverse that we won't worry about it.

from ooflib.SWIG.common import ooferror
from ooflib.common import debug
from ooflib.common import excepthook
import code
import compiler
import sys
import traceback

magicstring = 'O#O#F'    # used to construct fake file name
lenmagic = len(magicstring)

############

# We use the compiler module to find the smallest blocks of code that
# can be passed to the interpreter.  compiler.parseFile() returns Node
# objects, which usually, but not always (esp. in Python 2.3), have
# their lineno set.  We need the line number so that we know which
# lines to interpret and which to report in tracebacks and syntax
# errors.

def _findLineNo(node):
    # Find the minimum line number of a node and all of its child
    # nodes in the AST, skipping any children without line numbers.
    kidlines = [lineno
                for lineno in [_findLineNo(n) for n in node.getChildNodes()]
                if lineno is not None]
    if not kidlines:
        return node.lineno
    minkidline = min(kidlines)
    if node.lineno is None:
        return minkidline
    return min([node.lineno, minkidline])

############

class ScriptLoader(code.InteractiveInterpreter):
    def __init__(self, filename, locals=None):
        self.filename = filename
        # If the open is going to fail, we want it to happen as early
        # as possible.
        self.fileobj = open(self.filename)

        code.InteractiveInterpreter.__init__(self, locals)

        # Parse the file and find the python blocks.  We need to find
        # the blocks first, so that we don't accidentally execute an
        # "if" and strand its "else".
        modj = compiler.parseFile(self.filename)
        # modj is an AST.Module, modj.node is an AST.Stmt
        blocknodes = modj.node.nodes  # generic list of AST.Nodes
        # self.blocks is a list of line numbers where blocks start
        self.blocks = [_findLineNo(node) for node in blocknodes]

        self.currentline = 0
        self.error = 0
            
    def run(self):
        self.lines = self.fileobj.readlines()
        nlines = len(self.lines)
        nblocks = len(self.blocks)
        self.blocks.append(nlines+1)  # avoid special case for last block
        for blockno in range(nblocks):
            # line numbers start at 1, not 0
            blockstart = self.blocks[blockno] - 1
            blockend = self.blocks[blockno+1] - 1
            source = ''.join(self.lines[blockstart:blockend])
            more = self.runsource(source, magicstring+str(blockstart))
            if more:
                raise ooferror.ErrPyProgrammingError("Error reading script!")
            self.progress(blockend+1, nlines)
            if self.stop() or self.error:
                break
        self.done()
            
    def showtraceback(self):
        self.error = 1
        (type, value, tb) = sys.exc_info()
        tblist = traceback.extract_tb(tb)
        # python docs make obscure comments about avoiding circular
        # references to tracebacks.
        tb = None

        fixedtb = []
        for frame in tblist[1:]:  # frame = (filename, lineno, func, text)
            if frame[0][:lenmagic] == magicstring:
                lineno = int(frame[0][lenmagic:]) + frame[1]
                fixedtb.append((self.filename, lineno, frame[2],
                                self.lines[lineno-1][:-1]))
            else:                 # this frame is not from the file
                fixedtb.append((frame[0], frame[1], frame[2], frame[3]))
        # printTraceBack clears the exception in gui mode
        excepthook.printTraceBack(type, value, fixedtb) 

    def showsyntaxerror(self, filename=None):
        self.error = 1
        (etype, value, tb) = sys.exc_info()
        # The traceback information in tb is completely useless, but
        # fortunately it just so happens that we know what line of the
        # file we're on, so we don't need it.  All we really need is the
        # offset, which tells us where in the line the error occurred.
        tb = None
        if etype is SyntaxError:
            try:
                msg, (dummy_filename, lineno, offset, line) = value
            except:
                pass
            else:
                value = SyntaxError(msg, (self.filename, self.currentline,
                                          offset,
                                          self.lines[self.currentline-1]))
        excepthook.printTraceBack(etype, value, [])
             

    def progress(self, current, total):
        pass

    def stop(self):
        return 0

    def done(self):
        self.fileobj.close()
