# -*- python -*-
# $RCSfile: timedrivers.py,v $
# $Revision: 1.24.8.1 $
# $Author: langer $
# $Date: 2008-07-01 17:41:43 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


from ooflib.common import registeredclass
from ooflib.common import primitives
from ooflib.common.IO import parameter
from ooflib.SWIG.common import config
from ooflib.SWIG.common.IO import progressbar
from ooflib.SWIG.engine import steppingscheme
from ooflib.SWIG.common import switchboard
from ooflib.SWIG.engine import timestepadapter
from ooflib.engine import problem

import random, sys

class NeighborData:
    def __init__(self):
        self.connected = False
        self.data = {}
    
        
class TimeDriver (registeredclass.RegisteredClass):
    registry = []
    def redraw(self):
        switchboard.notify('redraw')
    def clean_up(self, solverdriver, subproblemcontext, report_success=1):
        ## TODO: get rid of subproblemcontext arg? solverdriver knows it.
        solverdriver.update_solver_state(report_success)
        # delete temporary stiffness matrix and associated objects
        solverdriver.clear_memory()


class Null(TimeDriver):
    tip = "Applies a linear or non-linear solver only once"
    def __init__(self):
        self.steppermethod = steppingscheme.NullScheme()

    def apply(self, solverdriver, subproblemcontext, verbosity = 1,
              report_success = 1):
        try:
            solverdriver.apply(self.steppermethod, subproblemcontext,
                               verbosity = verbosity)
        finally:
            self.clean_up(solverdriver, subproblemcontext, report_success)
            self.redraw()
    def apply_one_step(self, solverdriver, subproblemcontext, verbosity = 1,
                       report_success = 1):
        solverdriver.apply(self.steppermethod, subproblemcontext,
                           verbosity = verbosity)

if config.devel()>=1:

    registeredclass.Registration(
        'Null',
        TimeDriver, ## superclass
        Null, ## subclass
        ordering = 100,
        params = [],
        tip="Single application of matrix driver"
        )

    ## OLD WAY OF DOING TIME STEPPING STARTS HERE ##
    ##    def apply(self, solverdriver, meshcontext, verbosity = 1, report_success = 1):
    ##        while self.steppermethod.iterate(meshcontext) is not None: 
    ##            ## iterate while the looping condition is not met
    ##            ## first solve the problem as if it were time independent
    ##            solverdriver.apply(self.steppermethod, meshcontext, verbosity, report_success)
    ##            self.steppermethod.update(meshcontext) ## solution is updated here
    ##        solverdriver.update_solver_state(meshcontext, report_success)
    ##        solverdriver.clear_memory() ## deletes temporary stiffness matrix and associated objects
    ##        solverdriver.redraw()
    ## OLD WAY OF DOING TIME STEPPING ENDS HERE ##

    class Evolver(TimeDriver):
        tip = "For solving partial differential equations with implicit or explicit time dependence"
        def __init__(self, steppermethod, timeadaptermethod):
            self.steppermethod = steppermethod
            self.timeadaptermethod = timeadaptermethod

        def apply(self, solverdriver, subproblemcontext, verbosity = 1,
                  report_success = 1):
            while self.timeadaptermethod.iterate(subproblemcontext,
                                                 self.steppermethod): 
                ## iterate while the looping condition is not met
                self.timeadaptermethod.apply(self.steppermethod,
                                             solverdriver,
                                             subproblemcontext,
                                             verbosity,
                                             report_success)
                ## multiple steppers may be applied
                self.timeadaptermethod.update(subproblemcontext,
                                              self.steppermethod)
                ## tentative solution is accepted here, if ever:
                ## residual is calculated
                ## redraws, save to file, refine, checkpoints, all occur here

            ## cleanup of the core, and update of GUI related messages
            self.clean_up(solverdriver, subproblemcontext, report_success)
            self.redraw()

        def apply_one_step(self, solverdriver, subproblemcontext, verbosity = 1,
                           report_success = 1):
            self.timeadaptermethod.apply(self.steppermethod,
                                         solverdriver,
                                         subproblemcontext,
                                         verbosity,
                                         report_success)

    registeredclass.Registration(
        'Evolver',
        TimeDriver,
        Evolver,
        ordering=200,
        params=[
        parameter.RegisteredParameter('steppermethod',
                                      steppingscheme.TimeDependentPtr,
                                      tip ='algorithm to iterate over time'),
        parameter.RegisteredParameter('timeadaptermethod',
                                      timestepadapter.TimeStepAdapterPtr,
                                      tip ='mechanism to set time step when material is evolving in time')
        
        ],
        tip='For solving microstructural fields with implicit or explicit time dependence.'
        )
    



